from collections import namedtuple
import datetime
import logging
from typing import Any, Dict, Generator

import pytz
from jf_ingest import diagnostics, logging_helper
from jira import JIRA, JIRAError

from jf_ingest.utils import retry_for_429s
from jf_ingest.jf_jira.exceptions import NoAccessibleProjectsException

# jira renamed this between api versions for some reason
try:
    from jira.resources import AgileResource as AGILE_BASE_REST_PATH
except ImportError:
    from jira.resources import GreenHopperResource as AGILE_BASE_REST_PATH

logger = logging.getLogger(__name__)


def get_jira_connection(config, creds, max_retries=3) -> JIRA:
    kwargs = {
        "server": config.jira_url,
        "max_retries": max_retries,
        "options": {
            "agile_rest_path": AGILE_BASE_REST_PATH,
            "verify": not config.skip_ssl_verification,
        },
    }

    if creds.jira_username and creds.jira_password:
        kwargs["basic_auth"] = (creds.jira_username, creds.jira_password)
    elif creds.jira_bearer_token:
        kwargs["options"]["headers"] = {
            "Authorization": f"Bearer {creds.jira_bearer_token}",
            "Cache-Control": "no-cache",
            "Content-Type": "application/json",
            "Accept": "application/json",
            "X-Atlassian-Token": "no-check",
        }
    else:
        raise RuntimeError(
            "No valid Jira credentials found! Check your JIRA_USERNAME, JIRA_PASSWORD, or JIRA_BEARER_TOKEN environment variables."
        )

    jira_connection = JIRA(**kwargs)

    jira_connection._session.headers[
        "User-Agent"
    ] = f'jellyfish/1.0 ({jira_connection._session.headers["User-Agent"]})'

    return jira_connection


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_fields(
    jira_connection: JIRA,
    include_fields: list[str] = [],
    exclude_fields: list[str] = [],
) -> list[dict]:
    logger.info("downloading jira fields... ")

    filters = []
    if include_fields:
        filters.append(lambda field: field["id"] in include_fields)
    if exclude_fields:
        filters.append(lambda field: field["id"] not in exclude_fields)

    fields = [
        field
        for field in jira_connection.fields()
        if all(filter(field) for filter in filters)
    ]

    logger.info("✓")
    return fields


def _project_is_accessible(jira_connection: JIRA, project_id: str):
    try:
        retry_for_429s(
            jira_connection.search_issues, f"project = {project_id}", fields=["id"]
        )
        return True
    except JIRAError as e:
        # Handle zombie projects that appear in the project list
        # but are not actually accessible.
        if (
            e.status_code == 400
            and e.text
            == f"A value with ID '{project_id}' does not exist for the field 'project'."
        ):
            logging_helper.log_standard_error(
                logging.ERROR, msg_args=[project_id], error_code=2112,
            )
            return False
        else:
            raise


def _detect_project_rekeys_and_update_metadata(
    projects: list,
    jellyfish_project_ids_to_keys: dict[str, str],
    jellyfish_issue_metadata: dict[str, dict],
):
    rekeyed_projects = []
    for project in projects:
        # Detect if this project has potentially been rekeyed !
        if (
            project.id in jellyfish_project_ids_to_keys
            and project.raw["key"] != jellyfish_project_ids_to_keys[project.id]
        ):
            logger.debug(
                f'Project (project_id={project.id}) {project.raw["key"]} was detected as being rekeyed (it was previously {jellyfish_project_ids_to_keys[project.id]}. Attempting to re-download all related jira issue data'
            )
            rekeyed_projects.append(project.id)

    # Mark issues for redownload if they are associated with rekeyed projects
    for metadata in jellyfish_issue_metadata.values():
        if metadata["project_id"] in rekeyed_projects:
            # Updating the updated time for each issue will force a redownload
            metadata["updated"] = pytz.utc.localize(datetime.datetime.min).isoformat()


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_projects_and_versions(
    jira_connection: JIRA,
    jellyfish_project_ids_to_keys: dict[str, str],
    jellyfish_issue_metadata: str,
    include_projects: list[str],
    exclude_projects: list[str],
    include_categories: list[str],
    exclude_categories: list[str],
) -> list[dict]:
    logger.info("downloading jira projects... [!n]")

    filters = []
    if include_projects:
        filters.append(lambda proj: proj.key in include_projects)
    if exclude_projects:
        filters.append(lambda proj: proj.key not in exclude_projects)
    if include_categories:

        def _include_filter(proj):
            # If we have a category-based allowlist and the project
            # does not have a category, do not include it.
            if not hasattr(proj, "projectCategory"):
                return False

            return proj.projectCategory.name in include_categories

        filters.append(_include_filter)

    if exclude_categories:

        def _exclude_filter(proj):
            # If we have a category-based excludelist and the project
            # does not have a category, include it.
            if not hasattr(proj, "projectCategory"):
                return True

            return proj.projectCategory.name not in exclude_categories

        filters.append(_exclude_filter)

    all_projects = retry_for_429s(jira_connection.projects)

    projects = [
        proj
        for proj in all_projects
        if all(filt(proj) for filt in filters)
        and _project_is_accessible(jira_connection, proj.id)
    ]

    if not projects:
        raise NoAccessibleProjectsException(
            "No Jira projects found that meet all the provided filters for project and project category. Aborting... "
        )

    _detect_project_rekeys_and_update_metadata(
        projects=projects,
        jellyfish_project_ids_to_keys=jellyfish_project_ids_to_keys,
        jellyfish_issue_metadata=jellyfish_issue_metadata,
    )

    logger.info("✓")

    logger.info("downloading jira project components... [!n]")
    for p in projects:
        p.raw.update(
            {
                "components": [
                    c.raw for c in retry_for_429s(jira_connection.project_components, p)
                ]
            }
        )
    logger.info("✓")

    logger.info("downloading jira versions... [!n]")
    result = []
    for p in projects:
        versions = retry_for_429s(jira_connection.project_versions, p)
        p.raw.update({"versions": [v.raw for v in versions]})
        result.append(p.raw)
    logger.info("✓")
    return result


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_users(
    jira_connection,
    gdpr_active,
    quiet=False,
    required_email_domains=None,
    is_email_required=False,
) -> list[dict]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_resolutions(jira_connection: JIRA) -> list[dict]:
    logger.info("downloading jira resolutions... [!n]")
    result = [r.raw for r in retry_for_429s(jira_connection.resolutions)]
    logger.info("✓")
    return result


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_issuetypes(jira_connection, project_ids) -> list[dict]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_issuelinktypes(jira_connection: JIRA) -> list[dict]:
    """Download Jira Issue Link Types from the issueLinkType endpoint.

    Args:
        jira_connection (JIRA): A Jira connection, from the jira Python library

    Returns:
        list[dict]: A list of 'raw' JSON objects pulled directly from the issueLinkType endpoint
    """
    logger.info("downloading jira issue link types... [!n]")
    result = [lt.raw for lt in retry_for_429s(jira_connection.issue_link_types)]
    logger.info("✓")
    return result


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_priorities(jira_connection: JIRA) -> list[dict]:
    """Loads Jira Priorities from their API. Has 429 handling logic

    Args:
        jira_connection (JIRA): A Jira connection (with the provided Jira Library)

    Returns:
        list[dict]: A list of 'raw' JSON objects pulled from the 'priority' endpoint
    """
    logger.info("downloading jira priorities... [!n]")
    result = [p.raw for p in retry_for_429s(jira_connection.priorities)]
    logger.info("✓")
    return result


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_boards_and_sprints(
    jira_connection: JIRA, project_ids, download_sprints
) -> list[dict]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def get_issues(jira_connection, issue_jql, start_at, batch_size) -> list[dict]:
    return []


# TODO: Make this a dataclass. Not a fan of namedtuple
IssueMetadata = namedtuple("IssueMetadata", ("key", "updated"))


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_all_issue_metadata(
    jira_connection,
    all_project_ids,
    earliest_issue_dt,
    num_parallel_threads,
    issue_filter,
) -> Dict[int, IssueMetadata]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def detect_issues_needing_sync(
    issue_metadata_from_jira: Dict[int, IssueMetadata],
    issue_metadata_from_jellyfish: Dict[int, IssueMetadata],
) -> list[dict]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_worklogs(jira_connection, issue_ids, work_logs_pull_from) -> list[dict]:
    return []


# Returns an array of CustomFieldOption items
@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_customfieldoptions(jira_connection, project_ids) -> list[dict]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_statuses(jira_connection) -> list[dict]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def detect_issues_needing_re_download(
    downloaded_issue_id_and_key_tuples: set[tuple[str, str]],
    issue_metadata_from_jellyfish,
    issue_metadata_addl_from_jellyfish,
) -> list[dict]:
    return []


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def download_necessary_issues(
    jira_connection,
    issue_ids_to_download,
    include_fields,
    exclude_fields,
    num_parallel_threads,
    suggested_batch_size: int = 2000,
) -> Generator[Any, None, None]:
    return []



import dataclasses
import logging
import gzip
import json
import time
from typing import Any, Callable, Optional

from jira import JIRAError

from jf_ingest import logging_helper

logger = logging.getLogger(__name__)

RETRY_EXPONENT_BASE = 5


class RetryLimitExceeded(Exception):
    pass


class StrDefaultEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return str(o)


class IngestIOHelper:
    def __init__(self, s3_bucket: str, s3_path: str, local_file_path: str):
        self.s3_bucket = s3_bucket
        self.s3_path = (s3_path,)
        # EVERYTHING in this file path will (potentially) be uploaded to S3!
        # DO NOT put any creds file in this path!!!!
        self.local_file_path = local_file_path

    def _get_file_name(self, object_name: str, batch_number: Optional[int] = 0):
        return f'{object_name}{batch_number if batch_number else ""}.json.gz'

    def write_json_data_to_local(
        self,
        object_name: str,
        json_data: dict | list[dict],
        batch_number: Optional[int] = 0,
    ):
        try:
            file_name = self._get_file_name(
                object_name=object_name, batch_number=batch_number
            )
            full_file_path = f"{self.local_file_path}/{file_name}"
            logger.info(f"Attempting to save {object_name} data to {full_file_path}")
            with gzip.open(full_file_path, "w") as f:
                f.write(
                    json.dumps(json_data, indent=2, cls=StrDefaultEncoder).encode(
                        "utf-8"
                    )
                )
                logger.debug(
                    f"File: {full_file_path}, Size: {round(f.tell() / 1000000, 1)}MB"
                )
        except Exception as e:
            logger.error(
                f"Exception encountered when attempting to write data to local file! Error: {e}"
            )

    def upload_files_to_s3(self):
        # TODO: Write multi threaded uploader to upload this local_file_path
        raise NotImplementedError("Function not implemented!")


def get_wait_time(e: Optional[Exception], retries: int) -> int:
    """
    This function attempts to standardize determination of a wait time on a retryable failure.
    If the exception's response included a Retry-After header, respect it.
    If it does not, we do an exponential backoff - 5s, 25s, 125s.

    A possible future addition would be to add a jitter factor.
    This is a fairly standard practice but not clearly required for our situation.
    """
    # getattr with a default works on _any_ object, even None.
    # We expect that almost always e will be a JIRAError or a RequestException, so we will have a
    # response and it will have headers.
    # So I'm choosing to use the getattr call to handle the valid but infrequent possibility
    # that it may not (None or another Exception type that doesn't have a response), rather tha
    # preemptively checking.
    response = getattr(e, "response", None)
    headers = getattr(response, "headers", {})
    retry_after = headers.get("Retry-After")
    if retry_after is not None:
        return int(retry_after)
    else:
        return RETRY_EXPONENT_BASE ** retries


def retry_for_429s(f: Callable[..., Any], *args, max_retries: int = 5, **kwargs) -> Any:
    """
    This function allows for us to retry 429s from Jira. There are much more elegant ways of accomplishing
    this, but this is a quick and reasonable approach to doing so.

    Note:
        - max_retries=5 will give us a maximum wait time of 10m25s.
    """
    for retry in range(max_retries + 1):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            if (
                hasattr(e, "status_code")
                and e.status_code == 429
                and retry < max_retries
            ):
                wait_time = get_wait_time(e, retries=retry)
                logging_helper.log_standard_error(
                    logging.WARNING,
                    # NOTE: Getting the function name here isn't always useful,
                    # because sometimes we circumvent the JIRA standard library
                    # and use functions like "get" and "_get_json", but it's still
                    # better than nothing
                    msg_args=[f.__name__, retry, max_retries, wait_time],
                    error_code=3071,
                )
                time.sleep(wait_time)
                continue
            else:
                raise RetryLimitExceeded(e)



import pytest
from unittest.mock import patch
from jira import JIRAError
from jf_ingest.utils import (
    RETRY_EXPONENT_BASE,
    RetryLimitExceeded,
    get_wait_time,
    retry_for_429s,
)


MOCK_RETRY_AFTER_TIME = 12345


class MockJiraErrorResponse:
    headers = {"Retry-After": MOCK_RETRY_AFTER_TIME}


MOCKED_429_JIRA_ERROR = JIRAError(
    status_code=429,
    text="This is a spoofed Jira Error",
    response=MockJiraErrorResponse(),
)


def test_get_default_wait_time():
    # Get Default Wait Times
    for retry_num in range(0, 10):
        assert (
            get_wait_time(e=None, retries=retry_num) == RETRY_EXPONENT_BASE ** retry_num
        )


def test_get_wait_times_for_retry_error():
    # Test when we have a valid Error
    mock_error = MOCKED_429_JIRA_ERROR

    assert get_wait_time(mock_error, retries=0) == MOCK_RETRY_AFTER_TIME


def test_get_wait_time_for_non_retry_error():
    mock_non_retry_exception = Exception()
    retries = 5
    assert (
        get_wait_time(mock_non_retry_exception, retries=retries)
        == RETRY_EXPONENT_BASE ** retries
    )


def test_retry_for_429s_timeout():
    def always_raise_429():
        raise MOCKED_429_JIRA_ERROR

    with patch("jf_ingest.utils.time.sleep", return_value=0) as m:
        with pytest.raises(RetryLimitExceeded):
            retry_for_429s(always_raise_429)

        try:
            retry_for_429s(always_raise_429)
        except RetryLimitExceeded:
            m.assert_called_with(MOCK_RETRY_AFTER_TIME)


def test_retry_for_429s_retry_works():
    arg_dict = {"retry_count": 0}
    success_message = "Success!"

    def raise_429_three_times():
        if arg_dict["retry_count"] == 3:
            return success_message
        else:
            arg_dict["retry_count"] += 1
            raise MOCKED_429_JIRA_ERROR

    with patch("jf_ingest.utils.time.sleep", return_value=0) as m:
        assert retry_for_429s(raise_429_three_times) == success_message
        m.assert_called_with(MOCK_RETRY_AFTER_TIME)


def test_retry_for_429s_with_generic_error():
    def raise_non_429():
        raise Exception("Generic Error Testing")

    with pytest.raises(Exception):
        retry_for_429s(raise_non_429)



from collections import namedtuple
from datetime import datetime, timedelta
from functools import wraps
from io import TextIOWrapper
import json
import logging
import os
import pytz
import shutil
import subprocess
import threading
import time

import psutil


_DIAGNOSTICS_FILE: TextIOWrapper = None


def _diagnostics_file_closed():
    return isinstance(_DIAGNOSTICS_FILE, TextIOWrapper) and _DIAGNOSTICS_FILE.closed


"""
The diagnostics module is used to capture diagnostics data
for agent runs. When running for Managed Ingest we do not
need to use any of this logic. For unity's sake we will
run the shared functions (so far only capture_timing),
but output nothing
"""

logger = logging.getLogger(__name__)


def _write_diagnostic(obj):
    # If there is no diagnostics file to write to, do nothing
    if not _DIAGNOSTICS_FILE:
        logger.debug(f"No diagnostics file is set")
        return

    json.dump(obj, _DIAGNOSTICS_FILE)
    _DIAGNOSTICS_FILE.write("\n")  # facilitate parsing
    _DIAGNOSTICS_FILE.flush()


def capture_agent_version():
    git_head_hash = os.getenv("SHA")
    build_timestamp = os.getenv("BUILDTIME")
    _write_diagnostic(
        {"type": "agent_version", "sha": git_head_hash, "timestamp": build_timestamp}
    )


# NOTE: This is the only function that is called from within the jf_ingest module!
# (as of 8/2/2023 -Gavin)
def capture_timing(*args, **kwargs):
    def actual_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start_time = pytz.utc.localize(datetime.utcnow()).isoformat()
            if "func_name_override" in kwargs:
                func_name = kwargs.pop("func_name_override")
            else:
                func_name = func.__name__
            ret = func(*args, **kwargs)
            end_time = pytz.utc.localize(datetime.utcnow()).isoformat()
            diag_obj = {
                "type": "func_timing",
                "name": func_name,
                "start": start_time,
                "end": end_time,
            }
            if ret is not None:
                try:
                    diag_obj["num_items"] = len(ret)
                except TypeError:
                    if type(ret) is int:
                        diag_obj["num_items"] = ret
            _write_diagnostic(diag_obj)
            return ret

        return wrapper

    return actual_decorator


def capture_run_args(run_mode, config_file, outdir, prev_output_dir):
    _write_diagnostic(
        {
            "type": "run_args",
            "run_mode": run_mode,
            "config_file": config_file,
            "outdir": outdir,
            "prev_output_dir": prev_output_dir,
        }
    )


def capture_outdir_size(outdir):
    _write_diagnostic(
        {
            "type": "outdir_size",
            "size_kb": int(
                subprocess.check_output(["du", "-sk", outdir])
                .split()[0]
                .decode("utf-8")
            ),
        }
    )


def continually_gather_system_diagnostics(kill_event, outdir):
    def _flush_cached_readings(cached_readings):
        if not cached_readings:
            return
        _write_diagnostic(
            {
                "type": "sys_resources_60s",
                "start": pytz.utc.localize(cached_readings[0].time).isoformat(),
                "end": pytz.utc.localize(cached_readings[-1].time).isoformat(),
                "cpu_pct": ["%.2f" % r.cpu_pct for r in cached_readings],
                "mem_used_gb": ["%.2f" % r.memory_used_gb for r in cached_readings],
                "mem_pct": ["%.2f" % r.memory_pct for r in cached_readings],
                "disk_used_gb": ["%.2f" % r.disk_used_gb for r in cached_readings],
                "disk_pct": ["%.2f" % r.disk_pct for r in cached_readings],
            }
        )

    SysReading = namedtuple(
        "SysReading",
        ("time", "cpu_pct", "memory_used_gb", "memory_pct", "disk_used_gb", "disk_pct"),
    )

    now = datetime.utcnow()
    readings = threading.local()
    readings.cached_readings = []
    readings.last_reading_time = None
    readings.last_flush_time = now

    while True:
        if kill_event.is_set():
            _flush_cached_readings(readings.cached_readings)
            readings.cached_readings = []
            return
        else:
            now = datetime.utcnow()
            if not readings.last_reading_time or (
                now - readings.last_reading_time
            ) > timedelta(seconds=60):
                cpu = psutil.cpu_percent()
                memory = psutil.virtual_memory()
                disk = shutil.disk_usage(outdir)
                readings.cached_readings.append(
                    SysReading(
                        now,
                        cpu / 100,
                        (memory.total - memory.available) / (1024 ** 3),
                        (memory.total - memory.available) / memory.total,
                        disk.used / (1024 ** 3),
                        disk.used / disk.total,
                    )
                )
                readings.last_reading_time = now

            if now - readings.last_flush_time > timedelta(seconds=300):
                _flush_cached_readings(readings.cached_readings)
                readings.cached_readings = []
                readings.last_flush_time = now

            # Keep the sleep short so that the thread's responsive to the kill_event
            time.sleep(1)


def open_file(outdir):
    global _DIAGNOSTICS_FILE
    try:
        if _DIAGNOSTICS_FILE is None:
            _DIAGNOSTICS_FILE = open(os.path.join(outdir, "diagnostics.json"), "a")
        elif _diagnostics_file_closed():
            _DIAGNOSTICS_FILE = open(os.path.join(outdir, "diagnostics.json"), "a")
        else:
            logger.debug(f"Diagnostics file is already open")
    except Exception as e:
        # These messages are left at the DEBUG level for a better UX in agent
        logger.debug(
            f"Error opening Diagnostics file. Diagnostics will not be written. Error: {e}"
        )


def close_file():
    try:
        global _DIAGNOSTICS_FILE
        if _DIAGNOSTICS_FILE is None:
            logger.debug("Diagnostics file is already closed")
        elif _diagnostics_file_closed():
            logger.debug("Diagnostics file is already closed")
        else:
            logger.debug("Closing diagnostics file")
            _DIAGNOSTICS_FILE.close()
        _DIAGNOSTICS_FILE = None
    except Exception as e:
        # These messages are left at the DEBUG level for a better UX in agent
        logger.debug(
            f"Error closing Diagnostics file. Diagnostics will not be written. Error: {e}"
        )



from enum import Enum
import logging
from dataclasses import dataclass
from datetime import datetime
from itertools import chain
import traceback
from jf_ingest import diagnostics, logging_helper
from jf_ingest.jf_jira.auth import JiraAuthConfig, get_jira_connection
from jf_ingest.jf_jira.downloaders import (
    IssueMetadata,
    detect_issues_needing_re_download,
    detect_issues_needing_sync,
    download_all_issue_metadata,
    download_boards_and_sprints,
    download_customfieldoptions,
    download_fields,
    download_issuelinktypes,
    download_issuetypes,
    download_necessary_issues,
    download_priorities,
    download_projects_and_versions,
    download_resolutions,
    download_statuses,
    download_users,
    download_worklogs,
)

from jf_ingest.utils import IngestIOHelper

logger = logging.getLogger(__name__)


@dataclass
class JiraIngestionConfig:
    s3_bucket: str
    s3_path: str
    local_file_path: str
    company_slug: str

    # Jira Auth Info
    auth_config: JiraAuthConfig

    # Fields information
    # NOTE: I assumed these are all strs
    include_fields: list[str]
    exclude_fields: list[str]

    # Projects information
    # NOTE: I assumed these are all strs
    include_projects: list[str]
    exclude_projects: list[str]
    include_project_categories: list[str]
    exclude_project_categories: list[str]

    # Boards/Sprints
    download_sprints: bool

    # Issues
    earliest_issue_dt: datetime
    issue_download_concurrent_threads: int
    issue_jql: str
    jellyfish_issue_metadata: dict
    jellyfish_project_ids_to_keys: dict

    # worklogs
    download_worklogs: bool
    # Potentially solidify this with the issues date, or pull from
    work_logs_pull_from: int


class JiraObject(Enum):
    JiraFields = "jira_fields"
    JiraProjectsAndVersions = "jira_projects_and_versions"
    JiraUsers = "jira_users"
    JiraResolutions = "jira_resolutions"
    JiraIssueTypes = "jira_issuetypes"
    JiraLinkTypes = "jira_linktypes"
    JiraPriorities = "jira_priorities"
    JiraBoards = "jira_boards"
    JiraSprints = "jira_sprints"
    JiraBoardSprintLinks = "jira_board_sprint_links"
    JiraIssues = "jira_issues"
    JiraIssuesRedownloaded = "jira_issues_re_downloaded"
    JiraIssuesIdsDownloaded = "jira_issue_ids_downloaded"
    JiraIssuesIdsDeleted = "jira_issue_ids_deleted"
    JiraWorklogs = "jira_worklogs"
    JiraCustomFieldOptions = "jira_customfieldoptions"
    JiraStatuses = "jira_statuses"


@diagnostics.capture_timing()
@logging_helper.log_entry_exit()
def load_and_push_jira_to_s3(config: JiraIngestionConfig):
    try:
        # For JF Ingest logic, logger should always be set at debug!
        # This is for better log management when deal with agent,
        # where we sometimes want to hide stack traces from clients,
        # but also NOT upload their sensitive information
        original_log_level = logger.level
        logger.setLevel(level=logging.DEBUG)

        jira_connection = get_jira_connection(config=config.auth_config)
        ingest_io_helper = IngestIOHelper(
            s3_bucket=config.s3_bucket,
            s3_path=config.s3_path,
            local_file_path=f"{config.local_file_path}/jira",
        )

        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraFields.value,
            json_data=(jira_connection, config.include_fields, config.exclude_fields),
        )

        projects_and_versions = download_projects_and_versions(
            jira_connection=jira_connection,
            jellyfish_project_ids_to_keys=config.jellyfish_project_ids_to_keys,
            jellyfish_issue_metadata=config.jellyfish_issue_metadata,
            include_projects=config.include_projects,
            exclude_projects=config.exclude_projects,
            include_categories=config.include_project_categories,
            exclude_categories=config.exclude_project_categories,
        )

        project_ids = {proj["id"] for proj in projects_and_versions}
        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraProjectsAndVersions.value,
            json_data=projects_and_versions,
        )

        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraUsers.value,
            json_data=download_users(
                jira_connection,
                config.gdpr_active,
                required_email_domains=config.required_email_domains,
                is_email_required=config.is_email_required,
            ),
        )
        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraResolutions.value,
            json_data=download_resolutions(jira_connection),
        )

        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraIssueTypes.value,
            json_data=download_issuetypes(jira_connection, project_ids),
        )
        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraLinkTypes.value,
            json_data=download_issuelinktypes(jira_connection),
        )
        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraPriorities.value,
            json_data=download_priorities(jira_connection),
        )

        boards, sprints, links = download_boards_and_sprints(
            jira_connection, project_ids, config.download_sprints
        )
        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraBoards.value, json_data=boards
        )
        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraSprints.value, json_data=sprints
        )
        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraBoardSprintLinks.value, json_data=links
        )

        issue_metadata_from_jira = download_all_issue_metadata(
            jira_connection,
            project_ids,
            config.earliest_issue_dt,
            config.issue_download_concurrent_threads,
            config.issue_jql,
        )

        issue_metadata_from_jellyfish = {
            int(issue_id): IssueMetadata(
                issue_info["key"],
                datetime.fromisoformat(
                    issue_info["updated"]
                ),  # already includes TZ info
            )
            for issue_id, issue_info in config.jellyfish_issue_metadata.items()
        }

        issue_metadata_addl_from_jellyfish = {
            int(issue_id): (
                issue_info.get("epic_link_field_issue_key"),
                issue_info.get("parent_field_issue_key"),
            )
            for issue_id, issue_info in config.jellyfish_issue_metadata.items()
        }

        (
            missing_issue_ids,
            _,
            out_of_date_issue_ids,
            deleted_issue_ids,
        ) = detect_issues_needing_sync(
            issue_metadata_from_jira, issue_metadata_from_jellyfish
        )

        issue_ids_to_download = list(missing_issue_ids.union(out_of_date_issue_ids))

        # TODO: Double check this approach makes sense....
        # This downloaded_issue_id_and_key_tuples could definitely be improved
        downloaded_issue_id_and_key_tuples = set()
        for batch_number, issue_batch in enumerate(
            download_necessary_issues(
                jira_connection,
                issue_ids_to_download,
                config.include_fields,
                config.exclude_fields,
                config.issue_download_concurrent_threads,
            ),
            start=0,
        ):
            for issue in issue_batch:
                downloaded_issue_id_and_key_tuples.add(issue["id"], issue["key"])
            ingest_io_helper.write_json_data_to_local(
                object_name=JiraObject.JiraIssues.value,
                json_data=issue_batch,
                batch_number=batch_number,
            )

        issue_ids_needing_re_download = detect_issues_needing_re_download(
            downloaded_issue_id_and_key_tuples,
            issue_metadata_from_jellyfish,
            issue_metadata_addl_from_jellyfish,
        )

        # TODO: Double check this approach makes sense....
        # This downloaded_issue_id_and_key_tuples could definitely be improved
        redownloaded_issue_id_and_key_tuples = set()
        for batch_number, issue_batch in enumerate(
            download_necessary_issues(
                jira_connection,
                list(issue_ids_needing_re_download),
                config.include_fields,
                config.exclude_fields,
                config.issue_download_concurrent_threads,
            ),
            start=0,
        ):
            for issue in issue_batch:
                redownloaded_issue_id_and_key_tuples.add(issue["id"], issue["key"])
            ingest_io_helper.write_json_data_to_local(
                object_name=JiraObject.JiraIssuesRedownloaded.value,
                json_data=issue_batch,
                batch_number=batch_number,
            )

        all_downloaded_issue_ids = [
            int(i[0])
            for i in chain(
                redownloaded_issue_id_and_key_tuples,
                redownloaded_issue_id_and_key_tuples,
            )
        ]

        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraIssuesIdsDownloaded.value,
            json_data=all_downloaded_issue_ids,
        )
        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraIssuesIdsDeleted.value,
            json_data=list(deleted_issue_ids),
        )

        if config.download_worklogs:
            ingest_io_helper.write_json_data_to_local(
                object_name=JiraObject.JiraWorklogs.value,
                json_data=download_worklogs(
                    jira_connection,
                    all_downloaded_issue_ids,
                    config.work_logs_pull_from,
                ),
            )

        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraCustomFieldOptions.value,
            json_data=download_customfieldoptions(jira_connection, project_ids),
        )

        ingest_io_helper.write_json_data_to_local(
            object_name=JiraObject.JiraStatuses.value,
            json_data=download_statuses(jira_connection),
        )

        ingest_io_helper.upload_files_to_s3()

        return True
    except Exception as e:
        logger.exception(f"Exception Encountered: {e}")
        logger.debug(traceback.format_exc())
        return False
    finally:
        # Reset log level
        logger.setLevel(level=original_log_level)



import json
import os

import requests_mock

from jf_ingest.jf_jira.auth import JiraAuthConfig, get_jira_connection


def get_jira_mock_connection():

    _MOCK_SERVER_INFO_RESP = (
        '{"baseUrl":"https://test-co.atlassian.net","version":"1001.0.0-SNAPSHOT",'
        '"versionNumbers":[1001,0,0],"deploymentType":"Cloud","buildNumber":100218,'
        '"buildDate":"2023-03-16T08:21:48.000-0400","serverTime":"2023-03-17T16:32:45.255-0400",'
        '"scmInfo":"9999999999999999999999999999999999999999","serverTitle":"JIRA",'
        '"defaultLocale":{"locale":"en_US"}} '
    )

    auth_config = JiraAuthConfig(
        url="https://test-co.atlassian.net/",
        personal_access_token="asdf",
        company_slug="test_co",
    )

    with requests_mock.Mocker() as m:
        _register_jira_uri(m, "serverInfo", f"{_MOCK_SERVER_INFO_RESP}")
        _register_jira_uri_with_file(m, "field", "api_responses/fields.json")
        jira_conn = get_jira_connection(config=auth_config, max_retries=1)

    return jira_conn


def get_fixture_file_data(fixture_path: str):
    with open(f"{os.path.dirname(__file__)}/fixtures/{fixture_path}", "r") as f:
        return f.read()


def get_jellyfish_issue_metadata():
    with open(
        f"{os.path.dirname(__file__)}/fixtures/jellyfish_issue_metadata.json", "r"
    ) as f:
        return json.loads(f.read())["issue_metadata"]


def _register_jira_uri_with_file(
    mock: requests_mock.Mocker, endpoint: str, fixture_path: str
):
    with open(f"{os.path.dirname(__file__)}/fixtures/{fixture_path}", "r") as f:
        _register_jira_uri(mock, endpoint, f.read())


def _register_jira_uri(mock: requests_mock.Mocker, endpoint: str, return_value: str):
    _MOCK_REST_BASE_URL = "https://test-co.atlassian.net/rest/api/2"

    mock.register_uri(
        "GET", f"{_MOCK_REST_BASE_URL}/{endpoint}", text=return_value,
    )
